/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-01-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"

void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.01; /* revolutions per second */
	float angle=glutGet(GLUT_ELAPSED_TIME)*0.001*rotate_rate*360;
	
	/* Build the pixel shader */
	static GLhandleARB prog=makeProgramObjectFromFiles(
		"raytrace_vertex.txt","raytrace.txt");
	glUseProgramObjectARB(prog);
	glFastUniform3fv(prog,"C",1,camera);
	
	
	// Upload the uniform variables for each object:
	for (int obj=0;obj<2;obj++) {
		glPushMatrix(); // keep each object independent
		glTranslatef(obj,0.0,0.0); // translated a bit...
		glRotatef(angle+obj*87,1,0,1); // ...and slowly rotating

		// Quadric surface's matrix:
		osl::mat4 A(0.0f);
		A[0][0]=1.0; // x^2 term
		A[1][1]=0.0; // y^2 term
		A[2][2]=0.0; // z^2 term
		A[2][3]=1.0; // z term
		A[3][3]=obj; // constant term
		
		// Object matrix
		osl::mat4 mv; // object -> world space
		glGetFloatv(GL_MODELVIEW_MATRIX,&mv[0][0]);
		osl::mat4 B=inverse(mv); // world -> object
		osl::mat4 BAB=transpose(B)*A*B; // see rule to modify quadric object

		// CAN'T use glFastUniform calls, because they cache the variable name.
		char Ustr[100];
		sprintf(Ustr,"A%d",obj);
		glUniformMatrix4fvARB(glGetUniformLocationARB(prog,Ustr),1,0,&BAB[0][0]);
		sprintf(Ustr,"reflectance%d",obj);
		glUniform3fvARB(glGetUniformLocationARB(prog,Ustr),1,vec3(1,obj*0.5,1-obj*0.5)); // purple
		glPopMatrix();
	}
	
	// Only draw one set of faces, to avoid drawing raytraced geometry twice.
	//   (front side may get clipped if you're inside the object, so draw back)
	glEnable(GL_CULL_FACE); glCullFace(GL_FRONT);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA); // premultiplied alpha
	
	// Draw one *HUGE* sphere; this one proxy geometry covers all our objects.
	//   Keep in mind the far clipping plane in determining the sphere radius
	glutSolidSphere(20.0,8,6); 
	
	glUseProgramObjectARB(0);
	
	// For line drawing, don't write to depth buffer
	glDepthMask(GL_FALSE);
	glEnable(GL_LINE_SMOOTH);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); // non-premultiplied blend
	glLineWidth(2.1);
	oglCameraAxes();
	
	/* Grid on the ground */
	glLineWidth(0.1);
	double res=10.0;
	glBegin(GL_LINES);
	glColor3f(0,0,0);
	for (float t=-res;t<=res;t+=1.0) {
		float z=0.0;
		glVertex3f(t,-res,z); // -Y
		glVertex3f(t,+res,z); // +Y
		glVertex3f(-res,t,z); // -X
		glVertex3f(+res,t,z); // +X
		
	}
	glEnd();
	
	
	/*
	glColor3f(1.0,0.7,0.3); // orange ball
	glLineWidth(2.0);
	glutWireSphere(1.3,20,30);
	*/
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Raytrace with Shadowing");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0,-3,1);
	
	glutMainLoop();
	return 0;
}
